@@ -19,7 +19,6 @@ from flask.ext.admin.contrib.fileadmin import UploadForm, NameForm
 from flask.ext.admin.model.helpers import get_mdict_item_or_list
 from flask.ext.admin.helpers import get_form_data, validate_form_on_submit
 from flask.ext.admin.form import BaseForm, rules, get_form_opts
-from flask.ext.admin.actions import action
 
 from flask.ext.admin.contrib.sqla.tools import is_inherited_primary_key, get_column_for_current_model, get_query_for_ids
 
@@ -34,6 +33,9 @@ from werkzeug import secure_filename
 
 from flask_login import login_required, current_user, logout_user
 
+from flask.ext.admin import tools
+from flask.ext.admin._compat import text_type
+
 from field import *
 
 import shutil
@@ -41,7 +43,29 @@ import shutil
 import shelf.model
 db = shelf.model.db
 
-from model import LocalizedString, RemoteFile, LocalizedText, Picture
+from ..user.base import shelf_login_required
+
+from model import LocalizedString, RemoteFile, LocalizedText, Picture, WorkflowMixin, OrderableMixin
+
+def action(name, text, confirmation=None, icon=None):
+    """
+        Use this decorator to expose actions that span more than one
+        entity (model, file, etc)
+
+        :param name:
+            Action name
+        :param text:
+            Action text.
+        :param confirmation:
+            Confirmation text. If not provided, action will be executed
+            unconditionally.
+    """
+    def wrap(f):
+        f._action = (name, text, confirmation)
+        f._icon = (name, icon)
+        return f
+
+    return wrap
 
 def issubmodel(cls, name):
     if not cls:
@@ -84,13 +108,13 @@ class DashboardView(AdminIndexView):
     	   return render_template("shelf/login/recover.html", form=RecoverForm(request.form))
 
     @expose('/logout/', methods=("GET",))
-    @login_required
+    @shelf_login_required
     def logout(self):
         logout_user()
         return redirect(request.args.get("next") or url_for("admin.signin"))
 
     @expose()
-    @login_required
+    @shelf_login_required
     def index(self):
         return self.render(self._template, user=current_user, widgets=[w.render() for w in self.widgets])
 
@@ -99,6 +123,20 @@ class ShelfModelView(sqla.ModelView):
     create_template = "shelf/model/create.html"
     edit_template = "shelf/model/edit.html"
 
+    def _get_default_order(self):
+        """
+            Return default sort order
+        """
+        if self.column_default_sort:
+            if isinstance(self.column_default_sort, tuple):
+                return self.column_default_sort
+            else:
+                return self.column_default_sort, False
+        elif issubclass(self.model, OrderableMixin):
+            return 'shelf_order', True
+
+        return None
+
     def create_blueprint(self, admin):
         bp = super(ShelfModelView, self).create_blueprint(admin)
         @bp.context_processor
@@ -106,6 +144,16 @@ class ShelfModelView(sqla.ModelView):
             return dict(user=current_user)
         return bp
 
+    def scaffold_list_columns(self):
+        columns = super(ShelfModelView, self).scaffold_list_columns()
+        res = []
+        for c in columns:
+            if not c.startswith("shelf_"):
+                res.append(c)
+        if "shelf_state" in columns:
+            res.append("shelf_state")
+        return res
+
     def _order_by(self, query, joins, sort_field, sort_desc):
         """
             Apply order_by to the query
@@ -206,9 +254,181 @@ class ShelfModelView(sqla.ModelView):
 
         return super(ShelfModelView, self).get_list_value(context, model, name)
 
+    def icon_from_action(self, name):
+        for curname, icon in self._icons:
+            if name == curname:
+                return icon
+        return None
+
+    def get_actions_list(self):
+        """
+            Return a list and a dictionary of allowed actions.
+        """
+        actions = []
+        actions_confirmation = {}
+
+        for act in self._actions:
+            name, text = act
+
+            if self.is_action_allowed(name):
+                actions.append((name, text_type(text), self.icon_from_action(name)))
+
+                confirmation = self._actions_data[name][2]
+                if confirmation:
+                    actions_confirmation[name] = text_type(confirmation)
+
+        return actions, actions_confirmation
+
+    def init_actions(self):
+        """
+            Initialize list of actions for the current administrative view.
+        """
+        self._actions = []
+        self._icons = []
+        self._actions_data = {}
+
+        for p in dir(self):
+            attr = tools.get_dict_attr(self, p)
+
+            if hasattr(attr, '_action'):
+                name, text, desc = attr._action
+                name, icon = attr._icon
+
+                if name == "reorder":
+                    if not issubclass(self.model, OrderableMixin):
+                        continue
+                elif name == "publish" or name == "review" or name == "draft":
+                    if not issubclass(self.model, WorkflowMixin):
+                        continue
+
+                self._actions.append((name, text))
+                self._icons.append((name, icon))
+
+                # TODO: Use namedtuple
+                # Reason why we need getattr here - what's in attr is not
+                # bound to the object.
+                self._actions_data[name] = (getattr(self, p), text, desc)
+
+    @action('publish',
+            lazy_gettext('Publish'),
+            lazy_gettext('Are you sure you want to publish selected models?'),
+            "thumbs-up")
+    def action_publish(self, ids):
+        try:
+            query = get_query_for_ids(self.get_query(), self.model, ids)
+
+            name = self.model.__name__
+
+            count = 0
+
+            for m in query.all():
+                m.shelf_state = "public"
+                count += 1
+
+            self.session.commit()
+
+            if count == 1:
+                flash('{} was successfully published.'.format(name))
+            else:
+                flash('{} {}s were successfully published.'.format(count, name.lower()))
+        except Exception as ex:
+            if self._debug:
+                raise
+
+            flash(gettext('Failed to publish models. %(error)s', error=str(ex)), 'error')
+
+    @action('review',
+            lazy_gettext('Review'),
+            lazy_gettext('Are you sure you want to put selected models for review?'),
+            "search")
+    def action_review(self, ids):
+        try:
+            query = get_query_for_ids(self.get_query(), self.model, ids)
+
+            name = self.model.__name__
+
+            count = 0
+
+            for m in query.all():
+                m.shelf_state = "private"
+                count += 1
+
+            self.session.commit()
+
+            if count == 1:
+                flash('{} was successfully put for review.'.format(name))
+            else:
+                flash('{} {}s were successfully put for review.'.format(count, name.lower()))
+        except Exception as ex:
+            if self._debug:
+                raise
+
+            flash(gettext('Failed to put models for review. %(error)s', error=str(ex)), 'error')
+
+    @action('draft',
+            lazy_gettext('Draft'),
+            lazy_gettext('Are you sure you want to unpublish selected models?'),
+            'file-text-o')
+    def action_draft(self, ids):
+        try:
+            query = get_query_for_ids(self.get_query(), self.model, ids)
+
+            name = self.model.__name__
+
+            count = 0
+
+            for m in query.all():
+                m.shelf_state = "draft"
+                count += 1
+
+            self.session.commit()
+
+            if count == 1:
+                flash('{} was successfully unpublished.'.format(name))
+            else:
+                flash('{} {}s were successfully unpublished.'.format(count, name.lower()))
+        except Exception as ex:
+            if self._debug:
+                raise
+
+            flash(gettext('Failed to unpublish models. %(error)s', error=str(ex)), 'error')
+
+    @action('reorder',
+            lazy_gettext('Change order'),
+            lazy_gettext('Are you sure you want to reorder selected models?'),
+            'list-ol')
+    def action_reorder(self, ids):
+        try:
+            count = 0
+            name = self.model.__name__
+            models = []
+            priorities = []
+
+            for i in range(len(ids)):
+                m = db.session.query(self.model).get(ids[i])
+                models.append(m)
+                priorities.append(m.shelf_order)
+                count += 1
+
+            priorities = sorted(priorities, reverse=True)
+
+            for i in range(len(ids)):
+                m = models[i]
+                m.shelf_order = priorities[i]
+
+            db.session.commit()
+
+            if count == 1:
+                flash('{} was successfully reordered.'.format(name))
+            else:
+                flash('{} {}s were successfully reordered.'.format(count, name.lower()))
+        except Exception, ex:
+            flash(gettext('Failed to sort models. %(error)s', error=str(ex)), 'error')
+
     @action('delete',
             lazy_gettext('Delete'),
-            lazy_gettext('Are you sure you want to delete selected models?'))
+            lazy_gettext('Are you sure you want to delete selected models?'),
+            "trash-o")
     def action_delete(self, ids):
         try:
             query = get_query_for_ids(self.get_query(), self.model, ids)
@@ -237,12 +457,108 @@ class ShelfModelView(sqla.ModelView):
             flash(gettext('Failed to delete models. %(error)s', error=str(ex)), 'error')
 
     @expose('/')
-    @login_required
+    @shelf_login_required
     def index_view(self):
-        return super(ShelfModelView, self).index_view()
+        """
+            List view
+        """
+        # Grab parameters from URL
+        page, sort_idx, sort_desc, search, filters = self._get_extra_args()
+
+        # Map column index to column name
+        sort_column = self._get_column_by_idx(sort_idx)
+        if sort_column is not None:
+            sort_column = sort_column[0]
+
+        # Get count and data
+        count, data = self.get_list(page, sort_column, sort_desc,
+                                    search, filters)
+
+        # Calculate number of pages
+        num_pages = count // self.page_size
+        if count % self.page_size != 0:
+            num_pages += 1
+
+        # Pregenerate filters
+        if self._filters:
+            filters_data = dict()
+
+            for idx, f in enumerate(self._filters):
+                flt_data = f.get_options(self)
+
+                if flt_data:
+                    filters_data[idx] = flt_data
+        else:
+            filters_data = None
+
+        # Various URL generation helpers
+        def pager_url(p):
+            # Do not add page number if it is first page
+            if p == 0:
+                p = None
+
+            return self._get_url('.index_view', p, sort_idx, sort_desc,
+                                 search, filters)
+
+        def sort_url(column, invert=False):
+            desc = None
+
+            if invert and not sort_desc:
+                desc = 1
+
+            return self._get_url('.index_view', page, column, desc,
+                                 search, filters)
+
+        # Actions
+        actions, actions_confirmation = self.get_actions_list()
+
+        return self.render(self.list_template,
+                               data=data,
+                               # List
+                               list_columns=self._list_columns,
+                               sortable_columns=self._sortable_columns,
+                               # Stuff
+                               enumerate=enumerate,
+                               get_pk_value=self.get_pk_value,
+                               get_value=self.get_list_value,
+                               return_url=self._get_url('.index_view',
+                                                        page,
+                                                        sort_idx,
+                                                        sort_desc,
+                                                        search,
+                                                        filters),
+                               # Pagination
+                               count=count,
+                               pager_url=pager_url,
+                               num_pages=num_pages,
+                               page=page,
+                               # Sorting
+                               sort_column=sort_idx,
+                               sort_desc=sort_desc,
+                               sort_url=sort_url,
+                               # Search
+                               search_supported=self._search_supported,
+                               clear_search_url=self._get_url('.index_view',
+                                                              None,
+                                                              sort_idx,
+                                                              sort_desc),
+                               search=search,
+                               # Filters
+                               filters=self._filters,
+                               filter_groups=self._filter_groups,
+                               filter_types=self._filter_types,
+                               filter_data=filters_data,
+                               active_filters=filters,
+
+                               has_workflow=issubclass(self.model, WorkflowMixin),
+                               has_order=issubclass(self.model, OrderableMixin),
+
+                               # Actions
+                               actions=actions,
+                               actions_confirmation=actions_confirmation)
 
     @expose('/new/', methods=('GET', 'POST'))
-    @login_required
+    @shelf_login_required
     def create_view(self):
         """
             Create model view
@@ -268,11 +584,13 @@ class ShelfModelView(sqla.ModelView):
                            form=form,
                            form_opts=get_form_opts(self),
                            form_rules=self._form_create_rules,
+                           has_workflow=issubclass(self.model, WorkflowMixin),
+                           has_order=issubclass(self.model, OrderableMixin),
                            return_url=return_url)
 
     
     @expose('/edit/', methods=('GET', 'POST'))
-    @login_required
+    @shelf_login_required
     def edit_view(self):
         """
             Edit model view
@@ -308,6 +626,8 @@ class ShelfModelView(sqla.ModelView):
                            form=form,
                            form_opts=get_form_opts(self),
                            form_rules=self._form_edit_rules,
+                           has_workflow=issubclass(self.model, WorkflowMixin),
+                           has_order=issubclass(self.model, OrderableMixin),
                            return_url=return_url)
 
     def scaffold_form(self):
@@ -331,6 +651,11 @@ class ShelfModelView(sqla.ModelView):
                         and not issubmodel(self.model, 'Picture'):
                     setattr(cls, k, PictureField())
 
+        if issubclass(self.model, WorkflowMixin):
+            cls.shelf_state = wtforms.HiddenField()
+        if issubclass(self.model, OrderableMixin):
+            cls.shelf_order = wtforms.HiddenField()
+
         return cls
 
 class Library(fileadmin.FileAdmin):
@@ -348,7 +673,7 @@ class Library(fileadmin.FileAdmin):
         return bp
 
     @expose('/asyncupload', methods=("POST",))
-    @login_required
+    @shelf_login_required
     def async_upload(self):
         mfile = request.form['file']
         mname = request.form['name']
@@ -390,7 +715,7 @@ class Library(fileadmin.FileAdmin):
 
 
     @expose('/delete/', methods=('POST',))
-    @login_required
+    @shelf_login_required
     def delete(self):
         """
             Delete view method
@@ -436,7 +761,7 @@ class Library(fileadmin.FileAdmin):
 
     @expose('/mkdir/', methods=('GET', 'POST'))
     @expose('/mkdir/<path:path>', methods=('GET', 'POST'))
-    @login_required
+    @shelf_login_required
     def mkdir(self, path=None):
         """
             Directory creation view method
@@ -472,7 +797,7 @@ class Library(fileadmin.FileAdmin):
 
 
     @expose('/rename/', methods=('GET', 'POST'))
-    @login_required
+    @shelf_login_required
     def rename(self):
         """
             Rename view method
@@ -523,7 +848,7 @@ class Library(fileadmin.FileAdmin):
 
     @expose('/upload/', methods=('GET', 'POST'))
     @expose('/upload/<path:path>', methods=('GET', 'POST'))
-    @login_required
+    @shelf_login_required
     def upload(self, path=None):
         """
             Upload view method
@@ -581,7 +906,7 @@ class Library(fileadmin.FileAdmin):
 
     @expose('/modal-icons/')
     @expose('/modal-icons/b/<path:path>')
-    @login_required
+    @shelf_login_required
     def modal_iconic_index(self, path=None):
         """
             Index view method
@@ -653,7 +978,7 @@ class Library(fileadmin.FileAdmin):
 
     @expose('/modal/')
     @expose('/modal/b/<path:path>')
-    @login_required
+    @shelf_login_required
     def modal_index(self, path=None):
         """
             Index view method
@@ -725,7 +1050,7 @@ class Library(fileadmin.FileAdmin):
 
     @expose('/icons/')
     @expose('/icons/b/<path:path>')
-    @login_required
+    @shelf_login_required
     def iconic_index(self, path=None):
         """
             Index view method
@@ -797,7 +1122,7 @@ class Library(fileadmin.FileAdmin):
 
     @expose('/')
     @expose('/b/<path:path>')
-    @login_required
+    @shelf_login_required
     def index(self, path=None):
         """
             Index view method
